@@ -3,6 +3,8 @@ import unittest
 from io import BytesIO
 
 import detools
+from detools.common import pack_size
+from detools.common import unpack_size_bytes
 
 
 class DetoolsTest(unittest.TestCase):
@@ -14,9 +16,9 @@ class DetoolsTest(unittest.TestCase):
                             **kwargs):
         fpatch = BytesIO()
 
-        with open(from_filename, 'rb') as fold:
-            with open(to_filename, 'rb') as fnew:
-                detools.create_patch(fold, fnew, fpatch, **kwargs)
+        with open(from_filename, 'rb') as ffrom:
+            with open(to_filename, 'rb') as fto:
+                detools.create_patch(ffrom, fto, fpatch, **kwargs)
 
         actual = fpatch.getvalue()
         # open(patch_filename, 'wb').write(actual)
@@ -31,21 +33,21 @@ class DetoolsTest(unittest.TestCase):
                            to_filename,
                            patch_filename,
                            **kwargs):
-        patch_type = kwargs.get('patch_type', 'normal')
+        patch_type = kwargs.get('patch_type', 'sequential')
 
-        if patch_type == 'normal':
-            fnew = BytesIO()
+        if patch_type in ['sequential', 'hdiffpatch']:
+            fto = BytesIO()
 
-            with open(from_filename, 'rb') as fold:
+            with open(from_filename, 'rb') as ffrom:
                 with open(patch_filename, 'rb') as fpatch:
-                    to_size = detools.apply_patch(fold, fpatch, fnew)
+                    to_size = detools.apply_patch(ffrom, fpatch, fto)
 
-            actual = fnew.getvalue()
+            actual = fto.getvalue()
         elif patch_type == 'in-place':
             memory_size = kwargs['memory_size']
 
-            with open(from_filename, 'rb') as fold:
-                data = fold.read()
+            with open(from_filename, 'rb') as ffrom:
+                data = ffrom.read()
 
             data += (memory_size - len(data)) * b'\xff'
             fmem = BytesIO(data)
@@ -55,22 +57,23 @@ class DetoolsTest(unittest.TestCase):
 
             actual = fmem.getvalue()[:to_size]
         elif patch_type == 'bsdiff':
-            fnew = BytesIO()
+            fto = BytesIO()
 
-            with open(from_filename, 'rb') as fold:
+            with open(from_filename, 'rb') as ffrom:
                 with open(patch_filename, 'rb') as fpatch:
-                    to_size = detools.apply_patch_bsdiff(fold, fpatch, fnew)
+                    to_size = detools.apply_patch_bsdiff(ffrom, fpatch, fto)
 
-            actual = fnew.getvalue()
+            actual = fto.getvalue()
         else:
             raise Exception(patch_type)
 
-        with open(to_filename, 'rb') as fnew:
-            expected = fnew.read()
+        with open(to_filename, 'rb') as fto:
+            expected = fto.read()
 
         # open('actual-to.bin', 'wb').write(actual)
 
         self.assertEqual(to_size, len(expected))
+        self.assertEqual(len(actual), len(expected))
         self.assertEqual(actual, expected)
 
     def assert_create_and_apply_patch(self,
@@ -562,9 +565,7 @@ class DetoolsTest(unittest.TestCase):
                 with self.assertRaises(detools.Error) as cm:
                     detools.apply_patch(fold, fpatch, fnew)
 
-                self.assertEqual(
-                    str(cm.exception),
-                    "Expected patch type 0, but got 7.")
+                self.assertEqual(str(cm.exception), "Bad patch type 7.")
 
     def test_create_patch_foo_bad_patch_type(self):
         fpatch = BytesIO()
@@ -574,7 +575,9 @@ class DetoolsTest(unittest.TestCase):
                 with self.assertRaises(detools.Error) as cm:
                     detools.create_patch(fold, fnew, fpatch, patch_type='bad')
 
-                self.assertEqual(str(cm.exception), "Bad patch type 'bad'.")
+                self.assertEqual(
+                    str(cm.exception),
+                    "Bad algorithm (bsdiff) and patch type (bad) combination.")
 
     def test_apply_patch_foo_bad_compression(self):
         fnew = BytesIO()
@@ -827,6 +830,76 @@ class DetoolsTest(unittest.TestCase):
             'bsdiff.patch',
             patch_type='bsdiff')
 
+    def test_create_and_apply_patch_foo_hdiffpatch(self):
+        self.assert_create_and_apply_patch('tests/files/foo/old',
+                                           'tests/files/foo/new',
+                                           'tests/files/foo/hdiffpatch.patch',
+                                           algorithm='hdiffpatch',
+                                           patch_type='hdiffpatch')
+
+    def test_create_and_apply_patch_foo_match_blocks_sequential_none(self):
+        self.assert_create_and_apply_patch(
+            'tests/files/foo/old',
+            'tests/files/foo/new',
+            'tests/files/foo/match-blocks-sequential-none.patch',
+            patch_type='sequential',
+            algorithm='match-blocks',
+            compression='none',
+            match_block_size=8)
+
+    def test_create_and_apply_patch_foo_match_blocks_sequential(self):
+        self.assert_create_and_apply_patch(
+            'tests/files/foo/old',
+            'tests/files/foo/new',
+            'tests/files/foo/match-blocks-sequential.patch',
+            patch_type='sequential',
+            algorithm='match-blocks',
+            match_block_size=8)
+
+    def test_create_and_apply_patch_random_match_blocks_sequential_none(self):
+        self.assert_create_and_apply_patch(
+            'tests/files/random/from.bin',
+            'tests/files/random/to.bin',
+            'tests/files/random/match-blocks-sequential-none.patch',
+            compression='none',
+            patch_type='sequential',
+            algorithm='match-blocks',
+            match_block_size=64)
+
+    def test_create_and_apply_patch_random_match_blocks_hdiffpatch(self):
+        self.assert_create_and_apply_patch(
+            'tests/files/random/from.bin',
+            'tests/files/random/to.bin',
+            'tests/files/random/match-blocks-hdiffpatch.patch',
+            patch_type='hdiffpatch',
+            algorithm='match-blocks',
+            match_block_size=64)
+
+    def test_create_and_apply_patch_random_bsdiff(self):
+        self.assert_create_and_apply_patch('tests/files/random/from.bin',
+                                           'tests/files/random/to.bin',
+                                           'tests/files/random/patch-bsdiff.bin',
+                                           patch_type='bsdiff')
+
+    def test_pack_unpack_size(self):
+        datas = [
+            (-16_000_000_000, b'\xc0\x80\xe5\x9aw'),
+            ( -5_000_000_000, b'\xc0\xc8\xaf\xa0%'),
+            ( -2_000_000_000, b'\xc0\xd0\xac\xf3\x0e'),
+            (       -100_000, b'\xe0\x9a\x0c'),
+            (             -5, b'E'),
+            (              0, b'\x00'),
+            (              5, b'\x05'),
+            (        100_000, b'\xa0\x9a\x0c'),
+            (  2_000_000_000, b'\x80\xd0\xac\xf3\x0e'),
+            (  5_000_000_000, b'\x80\xc8\xaf\xa0%'),
+            ( 16_000_000_000, b'\x80\x80\xe5\x9aw')
+        ]
+
+        for value, packed in datas:
+            self.assertEqual(pack_size(value), packed)
+            self.assertEqual(unpack_size_bytes(packed), value)
+
 
 # This file is not '__main__' when executed via 'python setup.py3
 # test'.
